#include <stdio.h>
#include <stdlib.h>
#include <windows.h> // Probably need to make this more flexible
#include <openssl/sha.h> // Using SHA256 as random oracle
#include <openssl/rsa.h> // Using RSA for generating the modulus
#include <openssl/engine.h>
#include <math.h>

#define RtlGenRandom SystemFunction036 // CSPRNG of windows

BOOLEAN NTAPI RtlGenRandom(PVOID RandomBuffer, ULONG RandomBufferLength); // Used to pull secure random bits on windows

#define TWO_128_MINUS_1 (*(unsigned __int128*) "\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff")

long double *table = NULL; // Precomputation table for rejection sampling
size_t table_size = 0; // Size of the precomputation table

void precompute(long double stddev) // Fills the precomputation table according to required standard deviation
{
	long double pi=acosl(-1.0L);
	long double norm = 0.0L;
	//long double tmp = 0.0L;
	for (size_t i=0; i < table_size; ++i)
	{
		norm += expl(- pi * powl(((long double)((long double)i-(table_size/2.0L))/stddev),2.0L));
		table[i] = norm;
	}
	for (size_t i=0; i < table_size; ++i)
	{
		table[i] /= norm;
	}
}

void small_gaussian(BIGNUM *value) // Base sampler using rejection sampling for discrete gaussian
{
	unsigned __int128 tmp;
	RtlGenRandom((PVOID) &tmp,(ULONG) sizeof(unsigned __int128));
	long double	r = (long double) tmp / (long double) TWO_128_MINUS_1;
  	size_t a = 0;
  	size_t b = table_size-1;
  	size_t c;
  	while (b-a > 1)
  	{
  		c = (a+b)/2;
  		if (r > table[c])
  			a = c;
  		else
  			b = c;
  	}
	BN_set_word(value, b);
}

void precompute_reject(long double stddev) // Fills the precomputation table according to required standard deviation
{
	long double pi=acosl(-1.0L);
	//long double norm = 0.0L;
	//long double tmp = 0.0L;
	for (size_t i=0; i < table_size; ++i)
	{
		table[i] = expl(- pi * powl(((long double)i/stddev),2.0L));
		//norm += table[i];
	}
	/*for (size_t i=0; i < table_size; ++i)
	{
		table[i] /= norm;
	}*/
}

void small_gaussian_reject(BIGNUM *value) // Base sampler using rejection sampling for discrete gaussian
{
	unsigned __int128 tmp;
	int x;
	//int i=0;
	long double threshold;
	long double pass;
  	do
  	{
  		//++i;
  		RtlGenRandom((PVOID) &tmp,(ULONG) sizeof(unsigned __int128));
  		x = tmp % (2*table_size-1) - table_size + 1;
  		RtlGenRandom((PVOID) &tmp,(ULONG) sizeof(unsigned __int128));
		pass = (long double) tmp / (long double) TWO_128_MINUS_1;
		if (x >= 0)
			threshold = table[x];
		else
			threshold = table[-x];
	}
	while (pass >= threshold); //|| i < 100);
/*	if (i!=100)
		printf("SUCCESS\n");
	else
		printf("FAIL\n");*/
	BN_set_word(value, x);
}

void sample_gaussian(BIGNUM *value, size_t level, BIGNUM **z) // Samples bigger gaussians using the algorithm described in Micciancio-Walter 2018
{
	if (level == 0)
	{
		small_gaussian(value);
		return;
	}
	BIGNUM *x1 = BN_new();
	BIGNUM *x2 = BN_new();
	sample_gaussian(x1, level-1, z);
	sample_gaussian(x2, level-1, z);
	BN_CTX *aux = BN_CTX_new();
	BN_mul(value, x1, z[level-1], aux);
	BN_sub(x1, z[level-1], BN_value_one());
	BN_mul(x2, x2, x1, aux);
	BN_add(value, value, x2);
	BN_CTX_free(aux);
	BN_clear_free(x1);
	BN_clear_free(x2);
}

size_t find_target(BIGNUM *target, BIGNUM *sqrstddev, size_t level, BIGNUM ***z) // Find the require depth of recursion to achieve standard deviation target, stores auxilliary informations for the sampler
{
	if (BN_cmp(sqrstddev, target) == 1)
	{
		*z = malloc(level * sizeof(*z));
		return level;
	}
	BIGNUM *zi = BN_new();
	BIGNUM *newsqrstddev = BN_new();

	int k = BN_num_bits(sqrstddev);
	BN_rshift(zi, sqrstddev, (k+9)/2);

	BN_CTX *aux = BN_CTX_new();
	BIGNUM *tmp = BN_new();

	BN_sqr(tmp, zi, aux);

	BIGNUM *tmp2 = BN_new();
	BN_sub(tmp2, zi, BN_value_one());
	BN_sqr(tmp2, tmp2, aux);
	BN_add(tmp, tmp, tmp2);
	BN_clear_free(tmp2);

	BN_mul(newsqrstddev, sqrstddev, tmp, aux);

	BN_CTX_free(aux);
	BN_clear_free(tmp);

	size_t res = find_target(target, newsqrstddev, level + 1, z);
	(*z)[level] = BN_dup(zi);

	BN_clear_free(newsqrstddev);
	return res;
}

int mod_inverse(BIGNUM *a, BIGNUM *b, BIGNUM *x, BIGNUM *y) // Extended euclidean algorithm to find multiplicative inverse modulo some number
{
	
	if (BN_is_zero(a))
	{
		BN_zero(x);
		BN_one(y);
		return 0;
	}
	
	BN_CTX *aux = BN_CTX_new();

	BIGNUM *x1 = BN_new();
	BIGNUM *y1 = BN_new();
	BIGNUM *d = BN_new();
	BIGNUM *r = BN_new();

	BN_div(d, r, b, a, aux);

	mod_inverse(r, a, x1, y1);

	BN_copy(y, x1);
	BN_mul(r, d, y, aux);
	BN_sub(x, y1, r);

	BN_clear_free(y1);
	BN_clear_free(d);
	BN_clear_free(r);
	BN_CTX_free(aux);

	return 0;
}

typedef BIGNUM *dcr_pp_t;//type of public parameters

typedef BIGNUM *dcr_ct_t;// type of ciphertexts

typedef BIGNUM *dcr_sk_t; // type of user secret keys

typedef dcr_sk_t *dcr_msk_t; // type of master secret keys

void dcr_pp_init(dcr_pp_t *pp) // used to allocate memory for public parameters
{
	*pp = BN_new();
}

void dcr_pp_clear(dcr_pp_t pp) // used to free memory used by public parameters
{
	BN_clear_free(pp);
}

void dcr_ct_init(dcr_ct_t *ct) // used to allocate memory for a ciphertext
{
	*ct = BN_new();
}

void dcr_ct_clear(dcr_ct_t ct) // used to free memory used by a ciphertext
{
	BN_clear_free(ct);
}

void dcr_sk_init(dcr_sk_t *sk) // used to allocate memory for a user secret key
{
	*sk = BN_new();
}

void dcr_sk_clear(dcr_sk_t sk) // used to free memory used by a user secret key
{
	BN_clear_free(sk);
}

void dcr_msk_init(dcr_msk_t *msk, size_t n) // used to allocate memory for a master secret key
{
	*msk = malloc (sizeof *msk * n);
	for(size_t i=0; i < n; ++i)
		dcr_sk_init((*msk)+i);
}

void dcr_msk_clear(dcr_msk_t msk, size_t n) // used to free memory used by a master secret key
{
	for(size_t i=0; i < n; ++i)
		dcr_sk_clear(msk[i]);	
	free(msk);
}

void dcr_pp_keygen(dcr_pp_t *pp, size_t size) // used to generate the public parameters
{
	RSA* rsa = RSA_new();
	BIGNUM *e = BN_new();
	BN_set_word(e, 3);

	RSA_generate_key_ex(rsa, size, e, NULL);
	
	const BIGNUM *n;
    RSA_get0_key(rsa, &n, NULL, NULL);

	*pp = BN_dup(n);
	RSA_free(rsa);
}

void dcr_keygen(dcr_sk_t *sk, BIGNUM *stddev) // used to generate the i-th part of a master secret key
{
	BN_CTX *aux = BN_CTX_new();
	BIGNUM *target = BN_new();
	BIGNUM *sqrstddev = BN_new();
	BIGNUM **z;
	BN_sqr(target, stddev, aux);
	BN_set_word(sqrstddev, 4096);
	
	size_t res = find_target(target, sqrstddev, 0, &z);
	
	BN_clear_free(sqrstddev);
	BN_clear_free(target);
	
	sample_gaussian(*sk, res, z);

	for (size_t i=0; i < res; ++i)
		BN_clear_free(z[i]);
	free(z);
	BN_CTX_free(aux);
}

void dcr_msk_keygen(dcr_msk_t *msk, BIGNUM *stddev, size_t n) // used to generate a master secret key
{
	for (size_t i=0; i < n; ++i)
		dcr_keygen((*msk)+i, stddev);
}

void dcr_keyder(dcr_sk_t *sky, dcr_msk_t msk, BIGNUM **y, dcr_pp_t pp, size_t n){ // used to generate a secret key associated to a vector y
	BN_CTX *aux = BN_CTX_new();
	BIGNUM *tmp = BN_new();

	BN_zero(*sky);
	for(size_t i=0; i < n; ++i)
	{
		BN_mul(tmp, y[i], msk[i], aux);
		BN_add(*sky, *sky, tmp);
	}

	BN_clear_free(tmp);
	BN_CTX_free(aux);
}

void dcr_encrypt(dcr_ct_t *ct, dcr_sk_t sk, dcr_pp_t pp, BIGNUM *x, unsigned char *tau) // used to generate a partial ciphertext encrypting x for tag tau with secret key sk
{
	BN_CTX *aux = BN_CTX_new();
	
	BIGNUM *modulo = BN_new();
	BN_mul(modulo, pp, pp, aux);
	
	size_t size = 1 + (SHA256_DIGEST_LENGTH - 1 + BN_num_bytes(modulo))/SHA256_DIGEST_LENGTH;
	unsigned char* H;
	H = malloc(size * SHA256_DIGEST_LENGTH);
	
	char* ctr = malloc(sizeof(size));
	SHA256_CTX h;
	
	for (size_t i=0; i < size; ++i)
	{
		sprintf(ctr,"%x",i);
		SHA256_Init(&h);
		SHA256_Update(&h, tau, strlen(tau));
		SHA256_Update(&h, ctr, strlen(ctr));
		SHA256_Final(H+i*SHA256_DIGEST_LENGTH, &h);
	}
	free(ctr);

	BIGNUM *mask = BN_bin2bn(H, size * SHA256_DIGEST_LENGTH, NULL);
	BN_nnmod(mask, mask, modulo, aux);
	free(H);

	BN_mod_exp(mask, mask, sk, modulo, aux);
	BN_mul(*ct, x, pp, aux);
	BN_add(*ct, *ct, BN_value_one());
	BN_mod_mul(*ct, *ct, mask, modulo, aux);

	BN_clear_free(mask);
	BN_clear_free(modulo);
	BN_CTX_free(aux);
}

void dcr_decrypt(BIGNUM *m, dcr_sk_t sk, dcr_pp_t pp, dcr_ct_t *ct, BIGNUM **y, unsigned char *tau, size_t n) // used to decrypt a vector of ciphertexts with a functional secret key associated to vector y
{
	BN_CTX *aux = BN_CTX_new();
	
	BIGNUM *modulo = BN_new();
	BN_mul(modulo, pp, pp, aux);

	BN_RECP_CTX *aux_recp = BN_RECP_CTX_new();
	BN_RECP_CTX_set(aux_recp, modulo, aux);
	
	size_t size = 1 + (SHA256_DIGEST_LENGTH - 1 + BN_num_bytes(modulo))/SHA256_DIGEST_LENGTH;
	unsigned char* H;
	H = malloc(size * SHA256_DIGEST_LENGTH);
	
	char* ctr = malloc(sizeof(size));
	SHA256_CTX h;
	
	for (size_t i=0; i < size; ++i)
	{
		sprintf(ctr,"%x",i);
		SHA256_Init(&h);
		SHA256_Update(&h, tau, strlen(tau));
		SHA256_Update(&h, ctr, strlen(ctr));
		SHA256_Final(H+i*SHA256_DIGEST_LENGTH, &h);
	}
	free(ctr);

	BIGNUM *mask = BN_bin2bn(H, size * SHA256_DIGEST_LENGTH, NULL);
	BN_nnmod(mask, mask, modulo, aux);
	free(H);

	BIGNUM *tmp = BN_new();
	BIGNUM *e = BN_new();
	
	mod_inverse(mask, modulo, e, tmp);

	BN_nnmod(mask, e, modulo, aux);

	BN_mod_exp(tmp, mask, sk, modulo, aux);
	
	for (size_t i=0; i < n; ++i)
	{
		BN_mod_exp(e, ct[i], y[i], modulo, aux);
		BN_mod_mul_reciprocal(tmp, tmp, e, aux_recp, aux);
	}

	BN_mod_sub(tmp, tmp, BN_value_one(), modulo, aux);
	BN_div(m, e, tmp, pp, aux);

	if (! BN_is_zero(e))
	{
		printf("\n\n\nWARNING - - - THERE HAS BEEN AN ERROR DURING DECRYPTION\n\n\n");
	}

	BN_clear_free(e);
	BN_clear_free(tmp);
	BN_clear_free(mask);
	BN_RECP_CTX_free(aux_recp);
	BN_clear_free(modulo);
	BN_CTX_free(aux);
}

int main(){
	size_t size = 4096;
	size_t n = 100;
	BIGNUM *stddev = BN_new();

	struct timespec start, end;
	double deltat;

	dcr_pp_t pp;
	dcr_pp_init(&pp);

	clock_gettime(CLOCK_REALTIME, &start);
	dcr_pp_keygen(&pp, size);
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = (end.tv_nsec-start.tv_nsec)/1000000000.0;
	printf("Public parameters generation : %f s\n",deltat);

	table_size = 1024;
	table = malloc(sizeof *table * table_size);
	precompute(64);
	
	BN_CTX *ctaux = BN_CTX_new();
	BN_sqr(stddev, pp, ctaux);
	BN_mul(stddev, stddev, pp, ctaux);
	BN_sqr(stddev, stddev, ctaux);
	BN_CTX_free(ctaux);

	dcr_msk_t msk;
	dcr_msk_init(&msk, n);
	
	clock_gettime(CLOCK_REALTIME, &start);
	dcr_msk_keygen(&msk, stddev, n);
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = (end.tv_sec-start.tv_sec)+(end.tv_nsec-start.tv_nsec)/1000000000.0;
	printf("Master secret key generation : %f s\n",deltat);

	BIGNUM *x[n];
	BIGNUM *y[n];
	for (size_t i=0; i < n; ++i)
	{
		x[i] = BN_new();
		BN_set_word(x[i], i);
		y[i] = BN_new();
		BN_set_word(y[i], i);
	}
	
	dcr_sk_t sky;
	dcr_sk_init(&sky);

	clock_gettime(CLOCK_REALTIME, &start);
	dcr_keyder(&sky, msk, y, pp, n);
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = (end.tv_sec-start.tv_sec)+(end.tv_nsec-start.tv_nsec)/1000000000.0;
	printf("User secret key derivation : %f s\n",deltat);

	dcr_ct_t ctx[n];
	for (size_t i=0; i < n; ++i)
		dcr_ct_init(ctx+i);

	clock_gettime(CLOCK_REALTIME, &start);
	for (size_t i=0; i < n; ++i)
		dcr_encrypt(ctx+i, msk[i], pp, x[i], "a");
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = ((end.tv_sec-start.tv_sec)+(end.tv_nsec-start.tv_nsec)/1000000000.0)/n;
	printf("Encryption : %f s\n",deltat);

	BIGNUM *m = BN_new();

	clock_gettime(CLOCK_REALTIME, &start);
	dcr_decrypt(m, sky, pp, ctx, y, "a", n);
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = (end.tv_sec-start.tv_sec)+(end.tv_nsec-start.tv_nsec)/1000000000.0;
	printf("Decryption : %f s\n",deltat);

	char* toprint;
	printf("decrypted\n");
	toprint=BN_bn2dec(m);
	printf(toprint);
	printf("\nthat was the value\n");

	BIGNUM *result = BN_new();
	BN_zero(result);
	BN_CTX *aux = BN_CTX_new();
	BIGNUM *tmp = BN_new();
	for (size_t i=0; i < n; ++i)
		{
			BN_mul(tmp, x[i], y[i], aux);
			BN_add(result, result, tmp);
		}
	BN_CTX_free(aux);
	BN_clear_free(tmp);

	printf("expected was\n");
	toprint=BN_bn2dec(result);
	printf(toprint);
	printf("\n");

	BN_clear_free(result);
	BN_clear_free(m);
	for (size_t i=0; i < n; ++i)
		dcr_ct_clear(ctx[i]);
	dcr_sk_clear(sky);
	dcr_msk_clear(msk, n);
	dcr_pp_clear(pp);

	return 0;
}