#include <stdio.h>
#include <stdlib.h>
#include <windows.h> // Probably need to make this more flexible
#include <openssl/sha.h> // Using SHA256 as random oracle
#include <openssl/ec.h> // Using Elliptic Curve tools from openssl
#include <math.h>

#define RtlGenRandom SystemFunction036 // CSPRNG of windows

BOOLEAN NTAPI RtlGenRandom(PVOID RandomBuffer, ULONG RandomBufferLength); // Used to pull secure random bits on windows


void discrete_log(BIGNUM *x, EC_POINT *h, EC_GROUP *pp)
{
	BN_CTX *aux = BN_CTX_new();
	const EC_POINT *g = EC_GROUP_get0_generator(pp);
	EC_POINT *tmp = EC_POINT_new(pp);
	BN_zero(x);
	while (EC_POINT_cmp(pp, h, tmp, aux))
	{
		EC_POINT_add(pp, tmp, tmp, g, aux);
		BN_add(x, x, BN_value_one());
	}
	EC_POINT_free(tmp);
	BN_CTX_free(aux);
}

typedef EC_GROUP *ddh_pp_t;//type of public parameters

typedef EC_POINT *ddh_ct_t;// type of ciphertexts

typedef BIGNUM **ddh_sk_t; // type of user secret keys

typedef ddh_sk_t *ddh_msk_t; // type of master secret keys

void ddh_pp_init(ddh_pp_t *pp) // used to allocate memory for public parameters
{
	*pp = EC_GROUP_new_by_curve_name(928); // using brainpoolP256t1 (at least that corresponds on my machine)
	BN_CTX *aux = BN_CTX_new();
	EC_GROUP_precompute_mult(*pp, aux);
}

void ddh_pp_clear(ddh_pp_t pp) // used to free memory used by public parameters
{
	EC_GROUP_free(pp);
}

void ddh_ct_init(ddh_ct_t *ct, ddh_pp_t pp) // used to allocate memory for a ciphertext
{
	*ct = EC_POINT_new(pp);
}

void ddh_ct_clear(ddh_ct_t ct) // used to free memory used by a ciphertext
{
	EC_POINT_free(ct);
}

void ddh_sk_init(ddh_sk_t *sk) // used to allocate memory for a user secret key
{
	*sk = malloc (sizeof (**sk) * 2);
	(*sk)[0] = BN_new();
	(*sk)[1] = BN_new();
}

void ddh_sk_clear(ddh_sk_t sk) // used to free memory used by a user secret key
{
	BN_clear_free(sk[0]);
	BN_clear_free(sk[1]);
	free(sk);
}

void ddh_msk_init(ddh_msk_t *msk, size_t n) // used to allocate memory for a master secret key
{
	*msk = malloc (sizeof (**msk) * n);
	for(size_t i=0; i < n; ++i)
		ddh_sk_init((*msk)+i);
}

void ddh_msk_clear(ddh_msk_t msk, size_t n) // used to free memory used by a master secret key
{
	for(size_t i=0; i < n; ++i)
		ddh_sk_clear(msk[i]);	
	free(msk);
}

void ddh_keygen(ddh_sk_t *sk, const BIGNUM *p) // used to generate the i-th part of a master secret key
{	
	BN_rand_range((*sk)[0], p);
	BN_rand_range((*sk)[1], p);
}

void ddh_msk_keygen(ddh_msk_t *msk, ddh_pp_t pp, size_t n) // used to generate a master secret key
{
	const BIGNUM *p = EC_GROUP_get0_order(pp);
	for (size_t i=0; i < n; ++i)
		ddh_keygen((*msk)+i, p);
}

void ddh_keyder(ddh_sk_t *sky, ddh_msk_t msk, BIGNUM **y, ddh_pp_t pp, size_t n){ // used to generate a secret key associated to a vector y
	const BIGNUM *p = EC_GROUP_get0_order(pp);
	BN_CTX *aux = BN_CTX_new();
	BIGNUM *tmp = BN_new();

	BN_zero((*sky)[0]);
	for(size_t i=0; i < n; ++i)
	{
		BN_mod_mul(tmp, y[i], msk[i][0], p, aux);
		BN_mod_add((*sky)[0], (*sky)[0], tmp, p, aux);
	}

	BN_zero((*sky)[1]);
	for(size_t i=0; i < n; ++i)
	{
		BN_mod_mul(tmp, y[i], msk[i][1], p, aux);
		BN_mod_add((*sky)[1], (*sky)[1], tmp, p, aux);
	}
	
	BN_clear_free(tmp);
	BN_CTX_free(aux);
}

void ddh_encrypt(ddh_ct_t *ct, ddh_sk_t sk, ddh_pp_t pp, BIGNUM *x, unsigned char *tau) // used to generate a partial ciphertext encrypting x for tag tau with secret key sk
{
	BN_CTX *aux = BN_CTX_new();
	const BIGNUM *p = EC_GROUP_get0_order(pp);
	BN_RECP_CTX *aux_recp = BN_RECP_CTX_new();
	BN_RECP_CTX_set(aux_recp, p, aux);
		
	size_t size = 1 + (SHA256_DIGEST_LENGTH - 1 + BN_num_bytes(p))/SHA256_DIGEST_LENGTH;
	unsigned char* H;
	H = malloc(size * SHA256_DIGEST_LENGTH);
	BIGNUM *mask[2];

	SHA256_CTX h;
	char* ctr = malloc(sizeof(size));
	
	for (size_t i=0; i < size; ++i)
	{
		sprintf(ctr,"%x",i);
		SHA256_Init(&h);
		SHA256_Update(&h, tau, strlen(tau));
		SHA256_Update(&h, "0", strlen("0"));
		SHA256_Update(&h, ctr, strlen(ctr));
		SHA256_Final(H+i*SHA256_DIGEST_LENGTH, &h);
	}
	

	mask[0] = BN_bin2bn(H, size * SHA256_DIGEST_LENGTH, NULL);
	BN_nnmod(mask[0], mask[0], p, aux);

	for (size_t i=0; i < size; ++i)
	{
		sprintf(ctr,"%x",i);
		SHA256_Init(&h);
		SHA256_Update(&h, tau, strlen(tau));
		SHA256_Update(&h, "1", strlen("1"));
		SHA256_Update(&h, ctr, strlen(ctr));
		SHA256_Final(H+i*SHA256_DIGEST_LENGTH, &h);
	}
	free(ctr);

	mask[1] = BN_bin2bn(H, size * SHA256_DIGEST_LENGTH, NULL);
	BN_nnmod(mask[1], mask[1], p, aux);
	
	free(H);

	
	BN_mod_mul_reciprocal(mask[0], mask[0], sk[0], aux_recp, aux);
	BN_mod_mul_reciprocal(mask[1], mask[1], sk[1], aux_recp, aux);

	BN_mod_add(mask[0], mask[0], mask[1], p, aux);
	BN_mod_add(mask[0], mask[0], x, p, aux);

	EC_POINT_mul(pp, *ct, mask[0], NULL, NULL, aux);

	BN_clear_free(mask[0]);
	BN_clear_free(mask[1]);
	BN_RECP_CTX_free(aux_recp);
	BN_CTX_free(aux);
}

void ddh_decrypt(BIGNUM *m, ddh_sk_t sk, ddh_pp_t pp, ddh_ct_t *ct, BIGNUM **y, unsigned char *tau, size_t n) // used to decrypt a vector of ciphertexts with a functional secret key associated to vector y
{
	BN_CTX *aux = BN_CTX_new();
	const BIGNUM *p = EC_GROUP_get0_order(pp);
	BN_RECP_CTX *aux_recp = BN_RECP_CTX_new();
	BN_RECP_CTX_set(aux_recp, p, aux);
	
	size_t size = 1 + (SHA256_DIGEST_LENGTH - 1 + BN_num_bytes(p))/SHA256_DIGEST_LENGTH;
	unsigned char* H;
	H = malloc(size * SHA256_DIGEST_LENGTH);
	BIGNUM *mask[2];

	SHA256_CTX h;
	char* ctr = malloc(sizeof(size));
	
	for (size_t i=0; i < size; ++i)
	{
		sprintf(ctr,"%x",i);
		SHA256_Init(&h);
		SHA256_Update(&h, tau, strlen(tau));
		SHA256_Update(&h, "0", strlen("0"));
		SHA256_Update(&h, ctr, strlen(ctr));
		SHA256_Final(H+i*SHA256_DIGEST_LENGTH, &h);
	}
	

	mask[0] = BN_bin2bn(H, size * SHA256_DIGEST_LENGTH, NULL);
	BN_nnmod(mask[0], mask[0], p, aux);

	for (size_t i=0; i < size; ++i)
	{
		sprintf(ctr,"%x",i);
		SHA256_Init(&h);
		SHA256_Update(&h, tau, strlen(tau));
		SHA256_Update(&h, "1", strlen("1"));
		SHA256_Update(&h, ctr, strlen(ctr));
		SHA256_Final(H+i*SHA256_DIGEST_LENGTH, &h);
	}
	free(ctr);

	mask[1] = BN_bin2bn(H, size * SHA256_DIGEST_LENGTH, NULL);
	BN_nnmod(mask[1], mask[1], p, aux);
	
	free(H);

	BN_mod_mul_reciprocal(mask[0], mask[0], sk[0], aux_recp, aux);
	BN_mod_mul_reciprocal(mask[1], mask[1], sk[1], aux_recp, aux);

	BN_mod_add(mask[0], mask[0], mask[1], p, aux);
	BN_zero(mask[1]);
	BN_mod_sub(mask[0], mask[1], mask[0], p, aux);

	EC_POINT *target = EC_POINT_new(pp);
	EC_POINTs_mul(pp, target, mask[0], n, (const EC_POINT **)ct, (const BIGNUM **)y, aux);

	discrete_log(m, target, pp);

	EC_POINT_free(target);
	BN_clear_free(mask[0]);
	BN_clear_free(mask[1]);
	BN_RECP_CTX_free(aux_recp);
	BN_CTX_free(aux);
}

int main(){
	size_t n = 100;

	struct timespec start, end;
	double deltat;

	ddh_pp_t pp;

	clock_gettime(CLOCK_REALTIME, &start);
	ddh_pp_init(&pp);
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = (end.tv_nsec-start.tv_nsec)/1000000000.0;
	printf("Public parameters generation : %f s\n",deltat);
	
	ddh_msk_t msk;
	ddh_msk_init(&msk, n);

	clock_gettime(CLOCK_REALTIME, &start);
	ddh_msk_keygen(&msk, pp, n);
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = (end.tv_sec-start.tv_sec)+(end.tv_nsec-start.tv_nsec)/1000000000.0;
	printf("Master secret key generation : %f s\n",deltat);

	BIGNUM *x[n];
	BIGNUM *y[n];
	for (size_t i=0; i < n; ++i)
	{
		x[i] = BN_new();
		BN_set_word(x[i], i);
		y[i] = BN_new();
		BN_set_word(y[i], i);
	}
	
	ddh_sk_t sky;
	ddh_sk_init(&sky);

	clock_gettime(CLOCK_REALTIME, &start);
	ddh_keyder(&sky, msk, y, pp, n);
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = (end.tv_sec-start.tv_sec)+(end.tv_nsec-start.tv_nsec)/1000000000.0;
	printf("User secret key derivation : %f s\n",deltat);

	ddh_ct_t ctx[n];
	for (size_t i=0; i < n; ++i)
		ddh_ct_init(ctx+i, pp);

	clock_gettime(CLOCK_REALTIME, &start);
	for (size_t i=0; i < n; ++i)
		ddh_encrypt(ctx+i, msk[i], pp, x[i], "a");
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = ((end.tv_sec-start.tv_sec)+(end.tv_nsec-start.tv_nsec)/1000000000.0)/n;
	printf("Encryption : %f s\n",deltat);

	BIGNUM *m = BN_new();

	clock_gettime(CLOCK_REALTIME, &start);
	ddh_decrypt(m, sky, pp, ctx, y, "a", n);
	clock_gettime(CLOCK_REALTIME, &end);
	deltat = (end.tv_sec-start.tv_sec)+(end.tv_nsec-start.tv_nsec)/1000000000.0;
	printf("Decryption : %f s\n",deltat);

	char* toprint;
	printf("decrypted\n");
	toprint=BN_bn2dec(m);
	printf(toprint);
	printf("\nthat was the value\n");

	BIGNUM *result = BN_new();
	BN_zero(result);
	BN_CTX *aux = BN_CTX_new();
	BIGNUM *tmp = BN_new();
	for (size_t i=0; i < n; ++i)
		{
			BN_mul(tmp, x[i], y[i], aux);
			BN_add(result, result, tmp);
		}
	BN_CTX_free(aux);
	BN_clear_free(tmp);

	printf("expected was\n");
	toprint=BN_bn2dec(result);
	printf(toprint);
	printf("\n");

	BN_clear_free(result);
	BN_clear_free(m);
	for (size_t i=0; i < n; ++i)
		ddh_ct_clear(ctx[i]);
	ddh_sk_clear(sky);
	ddh_msk_clear(msk, n);
	ddh_pp_clear(pp);

	return 0;
}